#include "Matrioska/MatrioskaHeader.hpp"
//#include "VTO.class.hpp"

/**
	Changelog:

	* Version 0.4:
		- Context-aware trackers
		- context trackers rember if they were transitioned to inliers. If they were not, they will never be reinitialized

	* Version 0.3:
		- textureless patch filtering

	* Version 0.2:
		- adaptive/multisize tracker patch size
		- new state: transition (when many trackers are disabled, reinitialize them with a transition state)
		- give weights to trackers with past good performance

	* Version 0.1:
		- base system design

	Todo:
		- ratio between inlier/outlier should not count context trackers

*/



using namespace std;


	// ============
	//  Parameters
	// ============

	int patchSize = 12;					// The single patch size (sizeXsize)	[default: 12]
	int patchEnlargeWindow = 12;		// Window for searching correspondences	[default: 12]
	int patchSpace = -1;				// Spaces between patches				[default: -1]
										// Note: patchSize + patchSpace must be > 1

	int transitionThreshold = 2;		// To transition to an inlier the tracker must have agreed for the past frames [default: 3]

	int displacementErrorThreshold = 7;	// Max error to transitioning from inlier to outlier [default: 5]

	string datasetName = "woman";

	// Other vars
	Mat frame;

	Ptr<ColorBinarization> CB;


class WeakTracker {
	public:
		
		Point2i tl;				// Top-Left corner
		Point2i prevTL;			// previous position point
		

		int state;				// 1 => inlier; 2 => outlier; 3 => transition (markov model state)

		Point2i displacement;	// Difference with the previous position

		int goodEstimate;		// Number of correctly tracker frames

		// Original position
		Point2i relativeTL;		// Relative position inside the bounding box
		int patchSize;			// Size of the patch

		bool context;			// Context support tracker or object

		int inlierCount;		// Number of times this tracker was transitioned to inlier state (this is a good indicator for good context location)

		float colorWeight;		// Number of foreground pixels inside the tracker

		Mat patch;

		WeakTracker(Point2i relativeTL, int size, bool context = false) {

			this->relativeTL = relativeTL;
			this->patchSize = size;
			this->context = context;

			this->inlierCount = 0;
		}


		void setStateInlier() {
			state = 1;

			this->inlierCount++;
		}

		void setStateOutlier() {
			state = 2;
		}

		void setStateTransition() {
			state = 3;
		}

		// Set the old position to current position, and update the position
		void updateTL(Point2i tl) {
			
			this->prevTL = this->tl;
			this->tl = tl;

			this->displacement = this->tl - this->prevTL;

		}

		// Initialize the tracker in its default position and set it's state
		void reinit(const Mat& currentFrame, const Rect& objectROI, int setState = 1) {

			this->goodEstimate=0;
			
			Point2i tl = objectROI.tl() + this->relativeTL;
			updateTL(tl);
			
			Rect patchROI( tl.x, tl.y, this->patchSize, this->patchSize );

			
			// Disable current patch if: (i) it's out of the image or (ii) it's out of the bounding box
			if (!rectInsideImageCheck(currentFrame,patchROI) /*|| !objectROI.contains(this->tl)*/) { 
				
				setStateOutlier();
				return;	

			} 
			
			// If this is a context tracker, it should be initialized to transition state
			if (this->context==true)
				setStateTransition();
			else
				this->state = setState;
			

			this->patch = currentFrame(patchROI).clone();

			// Check if the current patch has some edges
			/*
			Mat currentPatch = currentFrame(patchROI);
			uchar previous = currentPatch.at<uchar>(0,0);
			int diff = 0;
			for(int i=0;i<currentPatch.rows;i++)
				for(int j=0;j<currentPatch.cols;j++) {
					int current = currentPatch.at<uchar>(i,j);
					diff += abs(previous - current);
					previous = current;
				}
			int meanPixelDifference = 1;
			if ( diff < patchSize*patchSize*meanPixelDifference ) {
				setStateOutlier();
				return;
			} 
			//cout << "Tracker intensity difference: "<<diff<<", threshold: "<<patchSize*patchSize*meanPixelDifference<<endl;
			//showImage(frame);
			//waitKey(0);
			*/

			this->draw(frame);	



			

			
		}

		void draw(Mat& frame) {
			// Default red
			Scalar color(0,0,255);

			// Transition yellow
			if (state==3)
				color = Scalar(0,255,255);

			// Context blu
			if (this->context==true)
				color = Scalar(255,0,0);

			Rect patchROI( tl.x, tl.y, this->patchSize, this->patchSize );
			drawRect(frame,patchROI,color,1);
			//drawRectCenter(frame,patchROI,color,1);
		}

		// Given the current patch size, calc the best search window
		int getSearchWindowSize() {
			/*	
				int size = this->patchSize*3;
				return min(size,12);			// Max search window size should be fixed
			*/
		}

		Rect getPatchRect(){

			return Rect(tl.x, tl.y, patchSize, patchSize);
		}

};

//> WeakTracker toString, for debugging
ostream& operator<<(ostream& strm, const WeakTracker& a) {
	return strm << " [Tracker] tl: " << a.tl <<  
		", prevTL: "<<a.prevTL<<
		", displacement: "<<a.displacement<<
		", state: "<<a.state<<
		", context: "<<(int)a.context;
}

vector<WeakTracker> trackers;

// Add a new tracker
void addTracker(const Mat& currentFrame, const Rect& objectROI, const Point2i& relativeTL, const int& patchSize, bool context = false){
	

	trackers.push_back(WeakTracker(
		relativeTL,
		patchSize,
		context
	));

	trackers.back().reinit(currentFrame,objectROI);

}

// Initialize context-trackers outside the objectROI
void initContextTrackers(const Mat& currentFrame, const Rect& objectROI) {
	
	int contextTrackerSize = 9;

	int trackersRow = objectROI.height / contextTrackerSize;
	int trackersCol = objectROI.width / contextTrackerSize;


	for(int i=-2;i<trackersRow+2;++i)
		for(int j=-2;j<trackersCol+2;++j) {

			Point2i relativeTL( (int)j*(contextTrackerSize) ,(int)i*(contextTrackerSize) );

			// Ensure the tracker is outside the object roi
			if ( relativeTL.x<0 || relativeTL.y<0 || 
				(relativeTL.x+contextTrackerSize)>objectROI.width || (relativeTL.y+contextTrackerSize)>objectROI.height )
				addTracker(currentFrame,objectROI,relativeTL,contextTrackerSize,true);

		}
	

	//showImage(frame);
	//waitKey(0);
}

// Initilize a grid of trackers with fixed size
void initTrackers(const Mat& currentFrame, const Rect& objectROI ) {

	int trackersRow = objectROI.height / (patchSpace+patchSize);
	int trackersCol = objectROI.width / (patchSpace+patchSize);

	//trackersRow++;trackersCol++;

	cout << "Initialized with Fixed Trackers ["<<patchSize<<"x"<<patchSize<<"]"<<endl<<
		"TrackersRow: "<<trackersRow<<", trackersCol: "<<trackersCol<<", total: "<<trackersCol*trackersRow<<endl;
	
	trackers.clear();

	for(int i=0;i<trackersRow;++i)
		for(int j=0;j<trackersCol;++j) {
			
			Point2i relativeTL((int)j*(patchSpace+patchSize) ,(int)i*(patchSpace+patchSize));

			addTracker(currentFrame,objectROI,relativeTL,patchSize);

		}
}

// Initialize a grid of trackers with differnt fixed sizes
void initMultiSizeTrackers(const Mat& currentFrame, const Rect& objectROI) {


	Log(1) << "Initialized with MultiSize Trackers" << endl;

	// Space between tracker's patches
	int currentPatchSpace = patchSpace;

	int minRoiSide = min(objectROI.width,objectROI.height);

	vector<int> sizes;

	// Small object

	if ( minRoiSide<50 ) 
		sizes.push_back(7);

	
	sizes.push_back(12);
	sizes.push_back(19);

	// If the object is very big prevent casting too many trackers
	if ( minRoiSide>80 ) {
		cout << "Detected large object, removing trackers.."<<endl;

		currentPatchSpace=1;
		sizes.pop_back();	// removes 19x19 tracker
	}

	trackers.clear();
	for(auto& currentPatchSize : sizes) {

		if (
			currentPatchSize<3 ||					// Patch Size too small
			currentPatchSize+currentPatchSpace < 2	// Patches too overlapped
		) {
			continue;
		}

		int trackersRow = objectROI.height / (currentPatchSize+currentPatchSpace);
		int trackersCol = objectROI.width / (currentPatchSize+currentPatchSpace);


		Log(1) << " Size ["<<currentPatchSize<<"x"<<currentPatchSize<<"] TrackersRow: "<<trackersRow<<", trackersCol: "<<trackersCol<<", total: "<<trackersCol*trackersRow<<endl;
	
		for(int i=0;i<trackersRow;++i)
			for(int j=0;j<trackersCol;++j) {
			
				Point2i relativeTL( (int)j*(currentPatchSize+currentPatchSpace) ,(int)i*(currentPatchSize+currentPatchSpace) );
			
				addTracker(currentFrame,objectROI,relativeTL,currentPatchSize);

			}
	}

	//initContextTrackers(currentFrame,objectROI);
}


// Reinit all trackers with state outlier (from "outlier" to "transition")
void disabledTrackersToTransition(const Mat& currentFrame, const Rect& objectROI) {

	for(auto i=0;i<trackers.size();++i) {
		if (trackers[i].state == 2) {

			// Do not transition if a tracker is context and was never good
			if (trackers[i].context == true && trackers[i].inlierCount == 0)
				continue;

			int state = 3;
			// If we disabled transitionThreshold put it directly as inlier
			if (transitionThreshold==0)
				state = 1;

			trackers[i].reinit(currentFrame,objectROI,state);			// Set transition mode

			
		}
	}

}


float BDF(string datasetName="", bool fixed = false) {
	
	


	// === 
	// DATASET
	// ===

		
		GTManager myGT;
		VideoCapture cap;


		string queryPath = "F:/Value/Personal/Parthenope/Project/input/babenko/VOTChallenge/"+datasetName+"/%08d.jpg";
		string init = "F:/Value/Personal/Parthenope/Project/input/babenko/VOTChallenge/"+datasetName+"/groundtruth.txt";


		//string queryPath = "F:/Value/Personal/Parthenope/Project/input/babenko/Other/Filmato.mp4";
		//string init = "F:/Value/Personal/Parthenope/Project/input/babenko/Other/groundtruth.txt";

		//queryPath = "F:/Value/Personal/Parthenope/Project/input/babenko/TLD/03_pedestrian1/%05d.jpg";
		//init = "F:/Value/Personal/Parthenope/Project/input/babenko/TLD/03_pedestrian1/init.txt";
		//myGT.useTLBR = true;

		/*string queryPath = "G:/LTDT/08_volkswagen/%05d.jpg";
		string init = "G:/LTDT/08_volkswagen/gt_mkv.txt";
		myGT.useTLBR = true;
		*/

		
		cap.open(queryPath);
		myGT.open(init);

		
	
	// ==== end dataset



	//load region, images and prepare for output
    //VOT vot_io("region.txt", "images.txt", "output.txt");		// VOT
		

	Counter counters;
	
	Timer t;

	VideoMaker video;

	int frameCount = 0;
	Point2i lastDisplacement;
	Rect objectROI;
	Mat currentFrame, prevFrame, backProjection;
	while (1) {

		Mat originalFrame;
		cap >> originalFrame;
		originalFrame.copyTo(currentFrame);

		//if (vot_io.getNextImage(currentFrame) != 1)	break;		// VOT



		//> Check if there are other frames to process (with webcam you need to manually exit with ESC key)
		if (currentFrame.empty())
			break;


		currentFrame.copyTo(frame);
		if (currentFrame.channels() > 1) {
			cvtColor(currentFrame, currentFrame, COLOR_BGR2GRAY);
		}
		else
			cvtColor(frame, frame, CV_GRAY2RGB);						//> Frame should be colored for drawing functions


		frameCount++;


		Log(1) << endl << "Parsing frame #" << frameCount << endl;


		Rect gt = myGT.getNextRect();

		// Init
		if (frameCount == 1) {

			CB = new ColorBinarization(gt);
			CB->binarize(originalFrame);


			//Rect gt = vot_io.getInitRectangle();			// VOT
			//vot_io.outputBoundingBox(gt);					// VOT

			objectROI = gt;

			Log() << endl << "====== " << datasetName << ", GT: " << gt << endl;
			if (fixed)
				initTrackers(currentFrame, objectROI);
			else
				initMultiSizeTrackers(currentFrame, objectROI);
			Log() << "Total final trackers: " << trackers.size() << endl << endl;

			//showImage(frame);
			//waitKey(0);

			myGT.calcOverlap(objectROI);

			currentFrame.copyTo(prevFrame);
			continue;
		}


		CB->binarize(originalFrame);
		backProjection = CB->backProjection;


		//vector<Point2i> displacements;
		//displacements.resize( trackers.size() );

		//   X		  Y  => Bucket
		map<int, map<int, int> > hs;

		int maxVote = 0;
		Point2i bestDisplacement;

		// Revoting
		if (frameCount > 2) {
			//hs[lastDisplacement.x][lastDisplacement.y]+=3;	// Cast a new vote in same direction of the previous displacement
		}

		// Check if a tracker is inside this area
		Rect objectROIEnlarged = rectScale(objectROI, 1.6f);

		// Displacement estimation
		for (auto i = 0; i < trackers.size(); i++) {

			auto& t = trackers[i];	// Current Tracker


			// If the current state is equal to oulier
			if (t.state == 2) {		// Skip local motion estimation
				continue;
			}


			// The tracker is outside the object
			if (!objectROIEnlarged.contains(t.tl)) {
				t.setStateOutlier();
				continue;
			}


			// Current tracker search window
			Rect searchWindow(
				t.tl.x - patchEnlargeWindow,
				t.tl.y - patchEnlargeWindow,
				t.patchSize + (2 * patchEnlargeWindow),
				t.patchSize + (2 * patchEnlargeWindow)
			);

			// Check if the search window is inside the image
			//if (!rectInsideImageCheck(currentFrame, searchWindow))
				//continue;
			rectInsideImage(currentFrame, searchWindow);


			// Search Window
			//drawRect(frame,searchWindow,Scalar(255,0,0));
			//showImage(frame);
			//waitKey(0);

			Rect currentPatchROI = t.getPatchRect();			// Rect(t.tl.x, t.tl.y, t.patchSize, t.patchSize);
			Mat currentPatch = prevFrame(currentPatchROI);

			//Mat currentPatch = t.patch;

			Mat result;
			matchTemplate(currentFrame(searchWindow), currentPatch, result, CV_TM_CCORR_NORMED);

			double minVal, maxVal; 	Point minLoc, maxLoc, matchLoc;
			minMaxLoc(result, &minVal, &maxVal, &minLoc, &maxLoc, Mat());
			matchLoc = maxLoc;
			matchLoc += searchWindow.tl();

			// Disable this tracker if its NCC is less than
			if (maxVal < 0.50) {
				t.setStateOutlier();
				continue;
			}

			// Updating references and displacement

			t.updateTL(matchLoc);
			
			//t.patch = currentFrame(t.getPatchRect()).clone();

			// ==== End update

			//cout << t << endl;


			// If current tracker is an inlier (tracker with state "transition" should not count for global motion estimate)
			if (t.state == 1) {

				int voteWeight = 2;			// Default weight (for non context tracker)

				// Increase weights
				if (t.goodEstimate > 9) {

					// Calc the tracker weight
					voteWeight += t.goodEstimate / 10;			// Give more weight to current tracker if it had good performance in the past

					voteWeight = min(voteWeight, 7);			// Set an upper limit to prevent giving too much weight to a single tracker
				}


				
				
				int voteColor = t.colorWeight * 5;
				voteWeight += voteColor;
				/*cout << "previous positive: " << t.colorWeight << ", votes: " << voteColor << endl;
				Mat drawFrame = frame.clone();
				t.draw(drawFrame);
				showImage(drawFrame, "tracker");
				waitKey(0);*/
				
				// Weight for context-tracker fixed to 1
				if (t.context)
					int voteWeight = 1;

				// Add a weighted vote to the current bucket
				int currentVote = (hs[t.displacement.x][t.displacement.y] += voteWeight);

				// Update current best
				if (currentVote > maxVote) {
					maxVote = currentVote;
					bestDisplacement = t.displacement;
				}
			}

			//if ( objectROI.contains(t.tl+Point2i(1,1)) )
			t.draw(frame);	// Draw current tracker

			//showImage(frame); waitKey(0);
		}

		lastDisplacement = bestDisplacement;         // could be used for revoting
		//printDoubleMap<map<int,map<int,int>>>(hs);
		//cout << "Best Displacement: "<<bestDisplacement<<" with votes: "<<maxVote<<endl;

		vector<Point> trackersPoints;				// set of tl and br of all trackers (used to calc boundingRect)

		vector<vector<Point> > inlierPositions(2);   // previous and current tracker position

		int disabledTrackers = 0;
		// Update trackers state
		for (int i = 0; i<trackers.size(); i++) {

			auto& t = trackers[i];
			//cout << "["<<i<<"]"<< t << endl;

			// If current state is outlier no need to recheck it
			if (t.state != 2) {

				// % of positive pixels inside this patch (will be used to weight the next vote)
				Mat binarizedPatch = backProjection(t.getPatchRect());
				int positivePixels = countNonZero(binarizedPatch);
				t.colorWeight = (float)positivePixels / t.getPatchRect().area();


				Point2i displacementError = t.displacement - bestDisplacement;

				// If current tracker displacement doesn't agree with the global mostion estimate, set the tracker as outlier
				if (abs(displacementError.x)>displacementErrorThreshold || abs(displacementError.y) > displacementErrorThreshold) {

					//if (t.colorWeight<0.3f)
						t.setStateOutlier();
					/*else {


						Mat drawFraming = currentFrame.clone();
						t.draw(drawFraming);
						showImage(drawFraming, "Good tracker");
						cout << "saving tracker with: " << t.colorWeight << endl;


					}*/
					

				} else {
					t.goodEstimate++;

					


					// If the current tracker is transition
					// and dispacement agrees with the global mostion estimate, set it as inliner
					if (t.state == 3) {

						// Give a more restrictive threshold for context trackers
						if (t.context == true && t.goodEstimate >= transitionThreshold * 15)
							t.setStateInlier();


						if (t.context == false && t.goodEstimate >= transitionThreshold)
							t.setStateInlier();
					}


					if (t.goodEstimate > 3) {

						trackersPoints.push_back(t.getPatchRect().tl());
						trackersPoints.push_back(t.getPatchRect().br());

						// Add previous and current displacement
						inlierPositions[0].push_back(t.prevTL);
						inlierPositions[1].push_back(t.tl);

					}


				}
			}

			if (t.state == 2)
				disabledTrackers++;
		}

		/*
		Rect trackersRect;
		if (!trackersPoints.empty()) {
			//drawPoints(frame, trackersPoints, Scalar(255, 255, 255));

			trackersRect = boundingRect(trackersPoints);
			drawRect(frame, trackersRect, Scalar(255, 255, 255));

			//objectROI = trackersRect;
		}
		else
			trackersRect = objectROI;
		*/

		// Move the  bounding box accordingly to the best displacement and check its position
		rectUpdate(objectROI, bestDisplacement);

		
		// Apply scale factor
		/*float scale = scaleEstimation(inlierPositions[0], inlierPositions[1]);
		cout << "Current scale: " << scale << endl;
		cout << "Current rect: " << objectROI << endl;
		objectROI = rectScale(objectROI, scale);
		cout << "Scaled rect: " << objectROI << endl;*/
		


		rectInsideImage(currentFrame,objectROI);

		// Info about trackers
		Log(1) << "Trackers: "<<trackers.size()<<", inlier/transition: "
			<<(trackers.size()-disabledTrackers)<<", threshold: "<<trackers.size()*0.35<<endl;
		//cout << "BestDisplacement: "<<bestDisplacement<<", voted by: "<<maxVote<<" trackers"<<endl;

		// If % of all trackers are disabled, reinit them
		if ((trackers.size()-disabledTrackers)<trackers.size()*0.35) {				// If less than inlier trackers survived
			
			int minTransitionDelay = 5;
			// Check if the previous transition was not too close
			if ( frameCount < minTransitionDelay || counters.check("transition",frameCount,minTransitionDelay) ) {

				counters.set("transition",frameCount);
				disabledTrackersToTransition(currentFrame,objectROI);
			}

		}

		// Draw bounding box
		drawRect(frame,objectROI);

		showImage(frame);
		showImage(currentFrame, "current frame");

		waitKey(1);

		

		cv::swap(prevFrame, currentFrame);
		

		//video.writeFrame(frame);


		float overlap = myGT.calcOverlap(objectROI);
		//cout << "Frame overlap: "<<overlap<<endl;
		//vot_io.outputBoundingBox(objectROI);			// VOT

	}	// ===== end loop over video

	double sec = t.timerEnd();
	cout << "Frame per seconds: "<<frameCount/sec<<", frames: "<<frameCount<<endl;

	float meanAccuracy = myGT.printStats();
	
	
	cout <<"========="<<endl<<endl<<endl;



	/*
	cout << "tracker size: "<<trackers.size()<<endl;
	for(int i=0;i<trackers.size();i++) {
		cout << "Trackers["<<i<<"], state: "<<trackers[i].state<<", goodEstimate: "<<trackers[i].goodEstimate<<endl;
	}
	*/

	//cin.get();

	
	destroyAllWindows();
	return meanAccuracy;
}


void main() {

	Log::level = 1;


	//BDF();	//VOT
	//return;



	vector<string> dataset;
	dataset.push_back("bicycle");
	dataset.push_back("bolt");
	dataset.push_back("car");
	dataset.push_back("cup");
	dataset.push_back("david");
	dataset.push_back("diving");
	dataset.push_back("face");
	dataset.push_back("gymnastics");
	dataset.push_back("hand");
	dataset.push_back("iceskater");
	dataset.push_back("juice");
	dataset.push_back("jump");
	dataset.push_back("singer");
	dataset.push_back("sunshade");
	dataset.push_back("torus");
	dataset.push_back("woman");
	
	dataset.clear();
	//dataset.push_back("jump");dataset.push_back("torus");dataset.push_back("woman");
	dataset.push_back("torus");

	vector<float> datasetAccuracy;
	for(auto& name : dataset) {
		datasetAccuracy.push_back(BDF(name));
		waitKey(100);
	}

	
	
	// size testing
	/*for(int i=12;i<13;i++) {
		patchSize = i;
		BDF("bolt",true);
	}*/
	

	cout << endl<<"=========="<<endl<<"FINAL MEAN ACCURACY: "<<meanVector<float>(datasetAccuracy)<<", over "<<datasetAccuracy.size()<<" datasets"<<endl;
	for (auto i = 0; i < datasetAccuracy.size(); ++i){

		cout << "Dataset: " << dataset[i] << ": " << datasetAccuracy[i] << endl;
	}
	sp();

}

