/*
Copyright � 2013 Mario Edoardo Maresca
 
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:
 
The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.
 
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
*/


//> LHTBucket represents each bucket of the hough space
//> It contains informations about each training images that projected the query center inside this bucket
//> LHTBucket.votes represents the number of votes the bucket received
class LHTBucket {
public:
	unsigned votes;						//> The number of votes this bucket received
	vector<Point2f> queryCenters;		//> The voted centers of the target object inside the query image (the current frame)
	vector<float> angles;				//> The voted angles of the target object

	vector<Point2f> trainCenters;		//> The centers of training images (the online learned object's centers)


	Point2f meanQueryCenter;			//> Having voted more centers we need to calc a mean query center (of the target object)
	Point2f meanTrainCenter;			//> "
	float meanAngle;					//> "

	Size2f meanTrainSize;				//> "


	map<size_t, vector<vector<Point2f>>> imgMatches;	//> For each training image stores the pair of matched keypoint's points (used with calcScale)
		
	//> ------

	vector<Point2f> trainPoints;		//> Can be used to calc the homography
	vector<Point2f> queryPoints;		//> Can be used to calc the homography
	vector<Point2f> trainCorners;		//> Can be used to calc the homography

	map<size_t, unsigned> votesImgIdx;	//> Mainly used for debug. Can be used to know which training image were used most

	//> ================================
	//>				METHODS
	//> ================================
	LHTBucket() {
		votes = 0;
	}

	//> Add a vote inside this bucket. For each vote we store:
	//>		- The target object center (inside the current frame)
	//>		- The target object angle
	//>		- The center of all training images that casted a vote
	//>		- The point of the training image (keypoint)
	//>		- The point of the query image	(keypoint)
	unsigned addVote(const Point2f& queryCenter, const float& angle, const Point2f& trainCenter, 
		const Point2f& trainP, const Point2f& queryP, bool addPoints = true ) {

		votes++;
		queryCenters.push_back(queryCenter);
		angles.push_back(angle);
		trainCenters.push_back(trainCenter);


		//> When you use Revote you don't want to readd the previous points too
		if (addPoints) {
			trainPoints.push_back(trainP);
			queryPoints.push_back(queryP);
		}
		return votes;
	}

	//> For each train images stores the points of train and query (used with calcScale)
	void addImagePoints(int imgIdx, const Point2f& trainP, const Point2f& queryP  ){

		if (imgIdx<0) //> When it's -1 it's a revote count. So don't consider it!
			return;

		votesImgIdx[imgIdx]++;	//> For each image store the number of votes (used for debug)

		//> Resize the vector to contain both training and query points
		if ( imgMatches[imgIdx].size()<2 )
			imgMatches[imgIdx].resize(2);

		//> Store the current Match for each training image 
		imgMatches[imgIdx][0].push_back(trainP);	//> First entry refers to train
		imgMatches[imgIdx][1].push_back(queryP);	//> Second entry refers to query

	}

	//> Return the Rect of the target object
	Rect getRect() {
		RotatedRect rRoi(meanQueryCenter,meanTrainSize,meanAngle*-1);
		return rectFromRotatedRect(rRoi);
	}

	//> Can be used for homography
	vector<Point2f>& getTrainCorners() {
		trainCorners.clear();

		trainCorners.push_back(Point2f( 0,0));
		trainCorners.push_back(Point2f( meanTrainCenter.x*2, 0 ));
		trainCorners.push_back(Point2f( meanTrainCenter.x*2,meanTrainCenter.y*2));
		trainCorners.push_back(Point2f( 0, meanTrainCenter.y*2 ));

		return trainCorners;
	}


};

//> LHTBucket toString, for debugging
ostream& operator<<(ostream& strm, const LHTBucket& a) {
	return strm << "LHTBucket, votes: " << a.votes << endl<< 
		"meanQueryCenter: "<<a.meanQueryCenter<<endl<< 
		"meanTrainSize: "<<a.meanTrainSize<<
		", meanTrainCenter: "<<a.meanTrainCenter<<endl<<
		"meanAngle: "<<a.meanAngle<<endl;
}

//> Localizer Hough Transform
class LocalizerHT {



protected:
	//> X        Y        Angle => entry.	 Access it with hs[10][20][35];
	map<int, map<int, map<int, LHTBucket> > > hs; //> HoughSpace

	int binSize;
	int binSizeAngle;
	unsigned minHoughVotes;
	unsigned revoteKeep;

	LHTBucket bestEntry;
	unsigned maxVotes;

	//> For debugging purpouse keep track of all queryCenter and angles
	vector<Point2f> allQueryCenter;
	vector<float> allAngles;

public:
	//> Assign default value
	LocalizerHT() {
		binSize = 35;
		minHoughVotes = 4;
		revoteKeep = 3;
		binSizeAngle = 25;
		maxVotes = 0;
	}
	LocalizerHT( int configBinSize, unsigned minVotes , unsigned revote) {
		binSize = configBinSize;
		minHoughVotes = minVotes;
		revoteKeep = revote;
		binSizeAngle = 25;
		maxVotes = 0;
	}

	//> Used for multiple keypoint-based methods and multiple training images
	void addMatches(const vector<DMatch>& match, 
					size_t engineIndex,			
					const vector<KeyPoint>& queryKP,
					const vector<MultiImage>& trains) {
		
		for( auto i = 0; i<match.size(); i++ ) {

			auto currentClass = match[i].imgIdx;
			auto qidx = match[i].queryIdx;
			auto tidx = match[i].trainIdx;

			KeyPoint train = trains[ currentClass ].keypoint[engineIndex][ tidx ];	//> keypoint on the training image
			KeyPoint query = queryKP[ qidx ];										//> keypoint on the query image

			LHTBucket currentEntry;
			Localizer l(trains[currentClass].image);
			l.useKeyPointMatch(train,query);


			addVote(l.queryCenter,l.angle,l.trainCenter,currentClass,currentEntry,train.pt,query.pt);

			//printMap<map<size_t,unsigned int>>(currentEntry.votesImgIdx);

			if (currentEntry.votes > maxVotes) {
				maxVotes = currentEntry.votes;
				bestEntry = currentEntry;
			}	

			if (maxVotes>3*minHoughVotes)
				break;
		}
	}

	//> Used for a single method only (and multiple training image)
	void addMatches(const vector<DMatch>& match,		
					const vector<KeyPoint>& queryKP,
					const vector<vector<KeyPoint>>& trainsKP,
					const vector<Mat>& trainsImage)
	{

		for( auto i = 0; i<match.size(); i++ ) {

			auto currentClass = match[i].imgIdx;
			if (trainsImage.size() == 1) 
				currentClass = 0;
			auto qidx = match[i].queryIdx;
			auto tidx = match[i].trainIdx;

			KeyPoint train = trainsKP[currentClass][ tidx ];					//> keypoint on the training image
			KeyPoint query = queryKP[ qidx ];									//> keypoint on the query image

			LHTBucket currentEntry;
			Localizer l(trainsImage[currentClass]);
			l.useKeyPointMatch(train,query);


			addVote(l.queryCenter,l.angle,l.trainCenter,currentClass,currentEntry,train.pt,query.pt);


			if (currentEntry.votes > maxVotes) {
				maxVotes = currentEntry.votes;
				bestEntry = currentEntry;
			}	

			if (maxVotes>3*minHoughVotes)
				break;
		}
	}

	//> Alias of addMatches. Use it with a single training image with a single method
	void addMatches(const vector<DMatch>& match,		
					const vector<KeyPoint>& queryKP,
					const vector<KeyPoint>& trainKP,
					const Mat& trainImage)
	{
		vector<vector<KeyPoint>> trainsKP;
		trainsKP.push_back(trainKP);

		vector<Mat> trainsImage;
		trainsImage.push_back(trainImage);

		addMatches(match,queryKP,trainsKP,trainsImage);
	}

	

	void addVote(const Point2f& queryCenter, const float& angle, const  Point2f& trainCenter, const  int& currentClass, 
		LHTBucket& entry, const  Point2f& trainP, const Point2f& queryP, bool addPoints = true) {
		int bucketX,bucketY,bucketAngle;

		//> Calc the right bucket
		bucketX = roundToNearest(queryCenter.x/binSize);
		bucketY = roundToNearest(queryCenter.y/binSize);
		bucketAngle = roundToNearest(angle/binSizeAngle);

		//> Adding vote
		hs[bucketX][bucketY][bucketAngle].addVote(queryCenter,angle,trainCenter,trainP,queryP,addPoints);
		hs[bucketX][bucketY][bucketAngle].addImagePoints( currentClass,trainP,queryP );




		//> Adding the vote for its around
		/*if (featureVoteAround) {
			int nBuckets = 1;
			for(int x=-nBuckets;x<=nBuckets;x++) {
				for(int y=-nBuckets;y<=nBuckets;y++) {
					if (x==0&&y==0)	//> The center vote was already added
						continue;

					LHTBucket &currentBucket = hs[bucketX + x][bucketY + y][bucketAngle];
					currentBucket.addVote(queryCenter,angle,trainCenter,trainP,queryP,addPoints);
					currentBucket.addImagePoints(currentClass,trainP,queryP);
					
				}
			}
		}*/

		entry = hs[bucketX][bucketY][bucketAngle];

		//> Adding for debugging
		//if (drawHoughVotes) {
			allQueryCenter.push_back(queryCenter);
			allAngles.push_back(angle);
		//}

		//cout << "Voting (#"<<entry.votes<<") to hs["<<bucketX<<"]["<<bucketY<<"]["<<bucketAngle<<"] imageIdx: ["<<currentClass<<"]"<<endl;


	}

	//> Add again the entry to the hough space
	void addPreviousVote(const LHTBucket& e) {
		
		//log("Adding previous entry with votes: "+toString(e.votes)+". With keep: "+toString(configFeatureRevoteKeep),2);


		//> Voting only "RevoteKeep" number of times
		LHTBucket fakeEntry; //> Needed only for addVote
		for(unsigned int i = 0;i<revoteKeep && i<e.votes;i++) {
			addVote( e.meanQueryCenter,e.meanAngle,e.meanTrainCenter,-1,fakeEntry,Point(0,0),Point(0,0),false);
		}

	}
	
	void getBestEntry(LHTBucket &currentBestEntry) {
		currentBestEntry = bestEntry;

		if (bestEntry.votes==0)
			return;

		Point2f meanQueryCenter;
		Point2f meanTrainCenter;
		float meanAngle;

		//> Calc mean center of query and train
		meanQueryCenter = meanPoint<Point2f>( bestEntry.queryCenters );
		meanTrainCenter = meanPoint<Point2f>( bestEntry.trainCenters );

		
		//> Calc mean angle
		float sumAngle = std::accumulate(bestEntry.angles.begin(),bestEntry.angles.end(),0.0f);
		meanAngle = sumAngle/bestEntry.votes;
		

		//> Save this value inside the entry
		currentBestEntry.meanQueryCenter = meanQueryCenter;
		currentBestEntry.meanTrainCenter = meanTrainCenter;
		currentBestEntry.meanAngle = meanAngle;

		//> Size
		currentBestEntry.meanTrainSize = Size2f(
			currentBestEntry.meanTrainCenter.x*2,
			currentBestEntry.meanTrainCenter.y*2
		);


	}



	void debug( Mat& queryImage ) {
		
		//drawGrid(queryImage,binSizeCoord);

		Mat houghSpace(queryImage.size(), CV_8U, Scalar::all(255));
		drawGrid(houghSpace,binSize);

		
		for(auto i=0;i<allQueryCenter.size();i++)
			drawCircle(houghSpace,allQueryCenter[i]);

		showImage(houghSpace,"Hough Space");
	}
};


class LHTBucket4{
	
	public:
		int votes;
		set<size_t> dmatchesIdx;
		vector<size_t> dmatchesIdx1;
		vector<size_t> dmatchesIdx2;
		vector<float> angles;
		vector<Point2f> queryCenters;
		vector<float> scales;

	LHTBucket4(){
		votes=0;

		dmatchesIdx1.reserve(40);
		dmatchesIdx2.reserve(40);
		angles.reserve(40);
		queryCenters.reserve(40);
		scales.reserve(40);


	}

};

// They key used for unordered_map of the accumulator
struct LHTBucket4Key {
    int shiftX, shiftY, angle, scale;

	LHTBucket4Key  (int _x, int _y, int _angle, int _scale) : shiftX (_x), shiftY (_y), angle (_angle), scale (_scale) {}

    bool operator==(LHTBucket4Key const& other) const {

        return std::tie(shiftX, shiftY, angle, scale) ==
               std::tie(other.shiftX, other.shiftY, other.angle, other.scale);

    }
};

template<typename T> void hashCombine(size_t& seed, T const& v){
	seed ^= hash<T>()(v) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
}

struct LHTBucket4KeyHash {

    std::size_t operator()(LHTBucket4Key const& i) const {
        std::size_t seed = 0;
        hashCombine(seed, i.shiftX);
        hashCombine(seed, i.shiftY);
        hashCombine(seed, i.angle);
        hashCombine(seed, i.scale);

        return seed;
    }
};