/*
	Copyright  2013 Mario Edoardo Maresca
 
	Permission is hereby granted, free of charge, to any person
	obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without
	restriction, including without limitation the rights to use,
	copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the
	Software is furnished to do so, subject to the following
	conditions:
 
	The above copyright notice and this permission notice shall be
	included in all copies or substantial portions of the Software.
 
	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
	HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
	OTHER DEALINGS IN THE SOFTWARE.
*/



class Matrioska : public Tracker {
public:

	//> ========================================================
	//>							 Properties
	//> ========================================================

		//> ====== Config Parameters ======

		//> Matrioska's sub-systems
		bool featureOnlinePositiveLearning;			//> Positive learning
		bool featureOnlineNegativeLearning;			//> Negative learning
			

		bool featureDiscardNegativeWhileLearning;			//>  Parse all new training keypoints and discard keypoints similar to the negative
		Size configFeatureDiscardNegativeWLenlargeROI;	//> (default 20,20) The area of the good object to obfuscate before learing a negative sample

		bool featureFallback;			//> Uses only keypoint engine sufficient for recongition (based on configHoughMinVotes)
										//> This setting doesn't affect if there is only one engine

		bool featureDiscardROITooFar;	//> If a new object is found but it's too far it is learned as a negative sample

		bool featureROI;				//> Limits the search only to a ROI
		Size configEnlargeROI;		//> Pixels to enlarge the ROI (should be set in according to the object speed)


		bool featureScale;				//> Estimate the object scale and adapt the bounding box

		//> Feature Pruning
		bool featurePruning;				//> Remove older training sample to avoid slow downs
		int configFeaturePruningWith;		//> Max number of keypoints before running pruning
		size_t configPruningStartFrom;		//> Don't delete the first training samples
			
		//> Feature online learning
		int configLearningPositiveStep;		//> Learn a positive sample each frame

		int configLearningNegativeStep;		//> Learn a negative sample each frame
		int configLearningNegativeAdiacentStep; //> Learn a negative adiacent roi each frame

		//> A lower ratio means less negative keypoints. 0.8% means 80% of negative keypoints and 20% positive
		float configFeatureLearningRatio;	//> The ratio between the negative keypoints and the total keypoints
											//> A lower value means less negative keypoints


		//> Hough-based localization
		bool featureVoteAround;				//> Each vote will cast a vote for its around (3x3)
		bool featureRevote;					//> Readd the previous good LHTBucket to the new hough space
		unsigned configHoughMinVotes;		//> Minimum number of votes to the same bucket to draw the bounding box
											//> It should be setted on the number of engines used

		unsigned configHoughRevoteKeep;		//> Revote "keep" number of times the previous good bucket if the object was found 
											//> (check featureRevote)
		

		int configHoughBinSize;				//> The width of the grid for hough voting



		int minObjectSide;			//> Width and Height can't be less than minObjectSide (this avoids errors with engines)

		float maxObjectShift;		//> When the object is found, the object in the next frame can't be farther than minObjectSize
									//> Note: set this according to the object's speed

		int limitGoodMatches;		//> Max good matches for each engine to be saved

		//> Grid Based option (used with ORB)
		int configDetectorCellSide;				//> Setting used with GridAdaptedFeatureDetector();
		int configDetectorKeyPointForCell;		//> Setting used with GridAdaptedFeatureDetector();
			
		//> ========== End Config ==========


		//> Debug/Draw settings
				
	
		int debugLevel;					//> Used by .log and .profile(Start|Debug) methods
		bool debugScale;				//> Shows pair of keypoints used by .calcScale method

		bool drawKP;						//> Draw keypoints
		bool drawKPNegativeDiscarded;		//> Draw the keypoints that passed the ratio test but were found on a negative sample

		bool drawHoughVotes;			//> Draw all votes done with hough voting
		bool drawTrainingImage;			//> Shows all training images with all keypoints

		bool writeOutputVideo;			//> Creates a video with the recognition
		VideoMaker outputVideo;


		//> Misc
		
		vector<KeyPointEngine> engine;
		
		


		//> Counter
		int frameWithObject;			//> Number of frames with a detection
		int addNegativeAt;				//> When reaches 0 add a new negative sample! (use it with negativeLearningStep)
		int previousGoodFrame;
		int consecutiveGoodFrames;

		int lastNegativePruning;		//> Used to bound the prune negative frequency

		
		clock_t start;
		

		//> Training pool
		int trainPositiveKeySample;		
		int trainNegativeSample;		//> Number of negative samples
		int trainTotalKeyPoint;			//> Positive+Negative keypoints in the training pool
		int trainNegativeKeyPoint;		//> Negative keypoints in the training pool

		//> Object
		vector<MultiImage> trains;		//> Store information about each training images
		

		Rect roi;

		LHTBucket objectHoughBucket;
		Size prevSize;
		Point2f topLeft;						//> Used for ROI

		
		
		
		Mat firstFrame;
		int newScaleFactor;			//> If the first target object is too small, we will resize the frame on-the-fly
		Mat detectionMask;

		//> Positive collage 
		int collageIndex;
		int collageToAddSize;
		int collageRows;
		Mat collage;

		//> Stats
		clock_t timeAll;
		vector<int> engineUsed; 
		
		

		vector<string> offlinePositiveSamples;
		vector<string> offlineNegativeSamples;



		// BDF
		Rect firstBBOriginal;
		Mat currentFrameOriginal;
		int nKeypoints = 0;
		bool useBDF = false;
		BDF trackerBDF;
		// ====================



	//> =======================
	//>     START METHODS
	//> =======================

	Matrioska(){
		init();
	
	}
	~Matrioska(){}

	//> Setter methods
	//> ==============

	//> Toggle debug windows and other minor debug information
	void setDebug(bool d=false) {

		if (!d) {
			debugLevel = 0;
			drawKP = false;
			debug = false;
		} else {
			debugLevel = 1;
			drawKP = true;
			debug = true;
		}
	}

	//> Train a positive sample from a file image
	void loadPositiveSample(const string& path) {

		offlinePositiveSamples.push_back(path);

	}

	//> Train a negative sample from a file image
	void loadNegativeSample(const string& path) {

		offlineNegativeSamples.push_back(path);

	}

	//> ================
	//>     PRESETS 
	//> ================

	//> You need to clear the default methods when you need to add your custom keypoint-based methods 
	//> with .addEngine
	void clearTechniquePool() {
		this->engine.clear();
	}

	//> Fast methods: ORB+FREAK
	void loadMethodsFast() {
		clearTechniquePool();
		addMethodORB();
		addMethodFREAK();
	}

	//> SURF alone
	void loadMethodsSingle() {
		clearTechniquePool();
		addMethodSURF();
	}

	//> Medium preset, overall better: ORB+SURF
	void loadMethodsMedium() {
		clearTechniquePool();

		addMethodORB();
		addMethodSURF();	
	}

	//> Slow but robust: ORB+SURF+SIFT
	void loadMethodsSlow() {
		clearTechniquePool();

		addMethodORB();
		addMethodSURF();
		addMethodSIFT();
	}



	void addMethodORB() {
		//> ORB: Blue
		addEngine(
			new OrbFeatureDetector(500,1.2f,8,14,0,2,0,14),   //> PatchSize from 31 to 14, for small images
			new OrbFeatureDetector(500,1.2f,8,14,0,2,0,14),
			new BFMatcher(NORM_HAMMING),
			"ORB",
			Scalar(255,0,0),
			1.0f,
			true
		);
	}
	void addMethodFREAK() {
		//> FREAK: Yellow
		addEngine(
			new OrbFeatureDetector(250,1.2f,8,14,0,2,0,14),		//> FREAK doesn't have a detector. Using ORB.
			new FREAK(true,true,4.0f,1),
			new BFMatcher(NORM_HAMMING),
			"FREAK",
			Scalar(0,255,255),
			0.95f
		);
	}
	void addMethodSURF() {
		//> SURF: Red
		addEngine(
			new SurfFeatureDetector(200,2,2),
			new SurfDescriptorExtractor(200,2,2),
			new BFMatcher(NORM_L2),
			"SURF",
			Scalar(0,0,255),
			1.0f,
			true
		);
	}
	void addMethodSIFT() {
		//> SIFT: Green
		addEngine(
			new SiftFeatureDetector(700,1),
			new SiftDescriptorExtractor(700,1),
			new BFMatcher(NORM_L2),
			"SIFT",
			Scalar(0,255,0),
			1.0f
		);
	}


	//> ==========================================
	//>              Algorithm
	//> ==========================================

	//> Tracks the object in the current frame
	void track() {



		//> Bootstrap and prepare for tracking
		this->prepare();

		clock_t startFrame = profileStart();
	
		//if (frameCount == 70)
			//configEnlargeROI=Point2f(60,40);

		//if (frameCount == 1 && firstFrame.size().width<360)
				//loadMethodsSingle();


		Log(1) << "\n\nFrame #" << frameCount << endl;
		
		Mat queryImage;							//> The query image (used to extract keypoints)
		if (frame.channels()>1)
			cvtColor(frame, queryImage, CV_RGB2GRAY);
		else {
			frame.copyTo(queryImage);
			cvtColor(frame,frame,CV_GRAY2RGB);	//> Frame should be colored for drawing functions
		}

		//equalizeHist(queryImage,queryImage);
		Mat fullQueryImage = queryImage;		//> The full image, needed when ROI is applied
		
		//> Initialization with first bouding box
		//> ====================================
		if (firstBB.area()>0 && frameCount == 1) {

			//> Ensure the first bounding box is not out of borders
			rectInsideImage(frame, roi);

			Log(1) << "First BB: "<<roi<<(newScaleFactor>1?" [ENLARGED!] ":"")<<endl;

			if (this->addTrainingSample(queryImage(roi))) {	

				for (auto e = 0; e < engine.size(); ++e){
					cout << "Engine: " << engine[e].name << " keypoints: " << trains[0].desc[e].rows << endl;
				}

				nKeypoints = trainTotalKeyPoint;
				cout << "Total positive keypoints: " << nKeypoints << endl;
				//sp();

				//this->addNegativeSample(roi, fullQueryImage);
				this->addNegativeAdiacent(fullQueryImage);
				
			}
			
			

			/*if (trainTotalKeyPoint-trainNegativeKeyPoint<3) {	//> too few keypoints extracted, use two technique
				loadMethodsSlow();
				trains.clear();
				if (this->addTrainingSample(queryImage(roi))) {
					this->addNegativeSample(roi, fullQueryImage);
				}
			}*/
			
			//debugShowImage(queryImage(roi), "Training Init");

			isObjectFound = true;
		}
		//> ===================================

		if (frameCount == 1) {
			trackerBDF.firstBB = this->firstBBOriginal;
			trackerBDF.corners = this->corners;
			trackerBDF.debug = this->debug;
		}

		trackerBDF.frame = currentFrameOriginal;
		trackerBDF.track();


		//> If Matrioska is likely to fail:
		//  (i) few keypoints (texture-less object)
		//  (ii) too many keypoints (background-clutter)
		//  (iii) failures within first frames
		//  Use BDF!
		if (nKeypoints<37 || nKeypoints>200 || useBDF) {
			//roi = ActiveTemplateMatching(frame, firstBB);
			//objectROI = getObjectROI();
			
			objectROI = trackerBDF.objectROI;

			isObjectFound = true;

			if (newScaleFactor > 1) {
				resize(frame, frame, Size(), (float)1.0f / newScaleFactor, (float)1.0f / newScaleFactor, INTER_NEAREST);
			}

			return;
		}

		if (trains.empty()) 
			Log::fatal("Empty train. Are you trying to track a texture-less object?");

		//> Start Matrioska algorithm
		//> ==========================

		//> Hough-based localization (see Paper)
		LocalizerHT Hough(configHoughBinSize,configHoughMinVotes,configHoughRevoteKeep);

		topLeft = Point2f(0,0);
		bool usingROI = false;
		//> If the we found the object in the previous frame
		if (isObjectFound) {

			//> Revote the previous good bucket in the hough space
			if (featureRevote)
				Hough.addPreviousVote(objectHoughBucket);

			//> Set a search window
			if (featureROI) {
				
				Rect searchWindow = rectSum(roi,configEnlargeROI);
				rectInsideImage(queryImage,searchWindow);

				if (rectSizeCheck(searchWindow,minObjectSide)) {

					usingROI = true;
					topLeft = searchWindow.tl();

					//> Extract the roi
					queryImage = queryImage(searchWindow);	//> The image to be parsed now it's only a sub-image.

					//> Draw search window (black)
					drawRect(frame,searchWindow,Scalar(0,0,0));
				}
				
			}
		
		} //> isObjectFound (Revote and ROI)

		//> Focus on the center if it's a big frame
		if (!usingROI && firstFrame.size().width>200) {
			
			// Central
			Rect roi = Rect(
				(int)(firstFrame.size().width * 0.1),	//> Cut left
				(int)(firstFrame.size().height * 0.1),	//> Cut top
				(int)(firstFrame.size().width * 0.8),	//> % of the width
				(int)(firstFrame.size().height * 0.75)	//> % of the height
			);


			queryImage = queryImage(roi);
			topLeft = roi.tl();
			usingROI = true;

			drawRect(frame,roi,(0,0,0));
		}


		vector<vector<KeyPoint>> queryKP;	//> For each engine save keypoints of the query
		Mat currentObjectROI;				//> The target object


		//> For each engine extract keypoints
		for(auto engineIndex=0;engineIndex<engine.size();engineIndex++) {

			//> Do not use more than one method on the entire frame
			//if ( featureFallback && !isObjectFound && engineIndex>0) 
				//break;

			//> STATS. Keep track of engines usage
			engineUsed[engineIndex]++;

			vector<KeyPoint> queryKPCurrent;
			Mat queryDescCurrent;  

			

			//> Extract keypoint to queryKP and descriptors to queryDesc using current engine
			this->feature(engine[engineIndex],queryImage,queryKPCurrent,queryDescCurrent);
			
			//> If we used a ROI restore original keypoint locations
			shiftKeyPoints(queryKPCurrent,topLeft);

			//> Store current KP
			queryKP.push_back(queryKPCurrent);

			//> Draw all keypoints just extracted on the query image (black)
			if (drawKP)
				drawKeypoints(frame,queryKPCurrent,frame,Scalar(30,30,30));

			vector<KeyPoint> queryInlierCandidates;	//> Used for drawing
			vector<DMatch> goodMatches;				//> The results of matrioska's matcher

			//> Match the extracted keypoints on the frame against the training pool (while discarding negative keypoints)
			this->match(engine[engineIndex], queryDescCurrent, goodMatches, 
							queryInlierCandidates, queryKPCurrent);

			//> Draw only the candidate good match keypoints
			if (drawKP)
				drawKeypoints(frame,queryInlierCandidates,frame,engine[engineIndex].color);

			//> Hough-based localization
			start = profileStart();
			Hough.addMatches(goodMatches,engineIndex,queryKPCurrent,trains);
			profileDebug(start,"Hough");

			//> If fallback is disabled always use all engines
			if (featureFallback==false && (engineIndex+1)<engine.size())
				continue;

			LHTBucket currentObjectBucket;
			Hough.getBestEntry(currentObjectBucket);

			Log(2) << "BestLHT entry votes: "  <<currentObjectBucket.votes<< ", min required: " << configHoughMinVotes<<endl;

			//> If the current LHTBucket has enough votes we have found the target object (a further check is done by isObjectCloseEnough)!
			if (currentObjectBucket.votes>=configHoughMinVotes) {

				//log("Number of votes for each training images");
				//printMap<map<size_t,unsigned int>>(currentObjectBucketvotesImgIdx);
		
				//> If featureScale is enabled run the scale estimation process
				if (featureScale && frameCount>20)
					this->calcScale(currentObjectBucket);

				if (currentObjectBucket.meanTrainSize.width<1 || currentObjectBucket.meanTrainSize.height<1) //> Something went wrong. Restore previous size
					currentObjectBucket.meanTrainSize = prevSize;

				//> Store the Rect of the target bounding box (already scaled by calcScale)
				roi = currentObjectBucket.getRect();

				int fullROIArea = roi.area();
				rectInsideImage(frame,roi);
				if (!rectSizeCheck(roi,minObjectSide)) { //> Check minObjectSide
					Log::warning("Object ROI too small: "+toString(roi));
					break;
				}
				int checkedROIArea = roi.area(); //> rectInsideImage can modify the Rect size
				if (fullROIArea>checkedROIArea)
					Log(2) << "Some of the object parts are out of scene" << endl;
				currentObjectROI = fullQueryImage(roi);

				//> ====== Check Target Object Distance ======
				if (!isObjectCloseEnough(currentObjectBucket,currentObjectROI))
					break;

						
				//> At this point the target object is found
				Log(2) << "Target Object FOUND!" << endl;

				//> Keep track of the usage of training images, for advanced pruning (not used)
				/*for( auto it = currentObjectBucketvotesImgIdx.begin(); it != currentObjectBucketvotesImgIdx.end(); ++it ) {
					//> Add only one for each train img idx
					trains[it->first].incrementUsage();
				}*/

				consecutiveGoodFrames++;
				frameWithObject++;

				previousGoodFrame = frameCount;
				isObjectFound = true;		
				objectHoughBucket = currentObjectBucket;
				prevSize = objectHoughBucket.meanTrainSize;

				//> ========== Online Positive/Negative Learning ======================
				this->onlineLearning(fullQueryImage,currentObjectROI,fullROIArea);
				//> ===================================================================

				//> if we are here, we don't need to use the next engine
				break; //> (go to end of "Engines")

			} //> Hough min votes
			
			//> We have another keypoint-based method to try
			if ( (engineIndex+1)<engine.size() ) 
				continue;

			//>  ============== End Hough-based localization


			//> If we are here the object was not found
			isObjectFound = false;
			consecutiveGoodFrames = 0;

		}	//> ===== Engines

		profileDebug(startFrame,"Frame total");


		//> Debugging Hough
		if (drawHoughVotes)
			Hough.debug(frame);

		this->debugTrainingPool();

		//>	============= Pruning ================
	
			//> Delete oldest 30% of training pool
			if ( featurePruning && trainTotalKeyPoint > configFeaturePruningWith ) {
				this->prune();
			}

			//> Too many negatives, try to keep the ratio defined by configFeatureLearningRatio
			if ( featurePruning && frameCount>30 && !areNegativesWithinLimit() ) {	
				this->pruneNegative();
			}

		//> ========================= End pruning

		
		//> Show the thumbnails of positive samples on top of the current frame
		collage.copyTo(frame(Rect(0,0,collage.size().width,collage.size().height)));

		if (isObjectFound) {
			objectROI = getObjectROI();
		}

		if (newScaleFactor > 1) {
			resize(frame, frame, Size(), (float)1.0f / newScaleFactor, (float)1.0f / newScaleFactor, INTER_NEAREST);
		}


		if (isObjectFound == false && frameCount < 21)
			useBDF = true;

		//> Next frame

	} //> End track()

	//> On some sequences the algorithm may automatically enlarge the frame size
	//> We need to scale the object ROI back to the original frame size if we want to compare it with the groundtruth
	//> The roi used during run-time is .roi
	Rect getObjectROI() {
		//> Shrink it
		Rect outputBB = Rect(roi.x+7,roi.y+7,roi.width-7,roi.height-7);

		//> Rescale back the bounding box
		outputBB = Rect(
			outputBB.tl().x / newScaleFactor,
			outputBB.tl().y / newScaleFactor,
			outputBB.width  / newScaleFactor,
			outputBB.height / newScaleFactor
		);

		return outputBB;
	}

	protected:
	//> Internal methods
	//> ================

		
		//> Non-config parameters default value  (should be called by constructors)
		void init() {

			//> Debug/draw setting
			debug = true;
			debugScale = false;
			debugLevel = 0;

			drawKP = true;						//> Draw keypoints
			drawKPNegativeDiscarded = false;	//> Draw keypoints that have passed the ratio test but were found on a negative samples

			drawHoughVotes = false;				//> Draw all GHT votes
			drawTrainingImage = false;			//> Show all training images with the keypoints detected

			writeOutputVideo = false;			//> Create a video with bounding box



			//> Stats about training pool
			trainPositiveKeySample = 0;
			trainNegativeSample = 0;
			trainTotalKeyPoint = 0;
			trainNegativeKeyPoint = 0;


			lastNegativePruning = 0;
			addNegativeAt = configLearningNegativeStep;
			newScaleFactor = 1;

		
		
			isObjectFound = false;				


			frameCount = 0;
			frameWithObject = 0;
			previousGoodFrame = 0;
			consecutiveGoodFrames = 0;

			timeAll = profileStart(0);


		
			


			//> Set default KeyPointEngine methods
			loadMethodsMedium();

			//> Config Parameters
			loadDefaultConfig();

		
		}

		//> Load default settings
		void loadDefaultConfig() {

			//> ==== Online Learning ====
			featureOnlinePositiveLearning = true;		//> Learning positive samples during tracking
			featureOnlineNegativeLearning = true;		//> Learn negative samples during tracking (the keypoints found outside the bounding box)
			
			featureDiscardNegativeWhileLearning = true;					//> Parse all new training keypoints and discard keypoints similar to the negative
			configFeatureDiscardNegativeWLenlargeROI = Size(10,10);	//> The area of the good object to obfuscate before learing a negative sample


			configLearningPositiveStep = 1;			//> Learn a positive sample each frame

			configLearningNegativeStep = 75;		//> Learn a negative sample each frame
			configLearningNegativeAdiacentStep = 10;	//> Learn a new adiacent negative region each frame

			configFeatureLearningRatio = 0.8f;		//> Negative keypoints should be % of total keypoints

			//> ==== Pruning ====
			featurePruning = true;					//> Remove older training sample to avoid slow downs
			configFeaturePruningWith = 25000;		//> Max number of keypoints before running pruning
			configPruningStartFrom = 15;				//> Don't delete the first training samples


			//> ==== Inter frame speed constraint ====
			featureDiscardROITooFar = true;			//> If a new object is found but it's too far from the previous detection: (a) (default) will not be learned as a new positive sample  OR 
																															// (b) will be learned as a new negative sample

			maxObjectShift = 200;					//> When the object is found, the object in the next frame can't be farther than minObjectSize
													//> Note: set this according to the object's speed

			//> ==== ROI ====
			featureROI = true;						//> Limit the search only to a ROI
			configEnlargeROI = Size(100,60);			//> Pixels to enlarge the ROI (should be set in according to the object speed)

			//> ==== HOUGH ====
			featureRevote = true;					//> Readd the previous good LHTBucket to the new hough space

			configHoughMinVotes = 4;				//> Minimum number of votes to the same bucket to draw the bounding box
													//> This setting can be auto tuned by Matrioska

			configHoughRevoteKeep = 3;				//> Revote "keep" number of times the previous good bucket if the object was found 
													//> (check featureRevote). It should set in according to configHoughMinVotes

			configHoughBinSize = 35;				//> The width of the cell for hough voting

			//> ==== Misc ====
			limitGoodMatches = 40;					//> Max good matches for each engine to be saved
			minObjectSide = 40;						//> Width and Height can't be less than minObjectSide (this avoids errors with engines)
			
			featureFallback = false;				//> Uses only keypoint engine sufficient for recongition (based on configHoughMinVotes)
													//> This setting doesn't affect if there is only one engine
			
			featureScale = false;					//> Scales the bounding box in according to ratio of keypoint pairs
			featureVoteAround = false;				//> Each vote will cast a vote for its around (3x3)

			configDetectorCellSide = 250;			//> Set the width and height of the grid cell (used with GridAdpatedFeatureDetector)
			configDetectorKeyPointForCell = 400;	//> Max number of keypoints inside the cell of size 250x250 (see configDetectorCellSide)	


		}

		//> Train images and set ground truth
		void prepare() {

			++frameCount;

			//> Run-once operation
			if (frameCount == 1) {
				
				firstBBOriginal = firstBB;

				//> Calc the scale factor, needed to enlarge the frame if the first object is too small
				//> ===================================================================================
				newScaleFactor = 1;
				if (firstBB.area()>0) {
					
					int minSide = min(firstBB.width,firstBB.height);
					if (minSide < 40) {

						newScaleFactor = (int)ceil(((float)40 / minSide));
						Log(0) << "Scale to be applied: "<<newScaleFactor<<endl;

						firstBB = Rect(
							firstBB.tl().x * newScaleFactor,
							firstBB.tl().y * newScaleFactor,
							firstBB.width  * newScaleFactor,
							firstBB.height * newScaleFactor
						);

						Log(1) << "Scaled first BB: " <<firstBB<< endl;

					}
					
					 //> We need to enlarge the first bounding box otherwise keypoints near the target object border will not be extracted
					roi = Rect(firstBB.x -7,firstBB.y-7,firstBB.width+7,firstBB.height+7);

				}
				//> ===================================================================================


				this->engineUsed.resize(50);		// reserve enough space for engine stats
			} 
			
			frame.copyTo(currentFrameOriginal);

			//> Applying scale factor (if the first bounding box was too small we must resize the whole frame)
			if (newScaleFactor > 1) {
				resize(frame, frame, Size(), newScaleFactor, newScaleFactor, CV_INTER_LINEAR);
			} 

			//> Prepare output video and collage (based on the size of the frame)
			if (frameCount==1) {
				firstFrame = frame;
				Log()<<"Frame size: " << firstFrame.size()<<endl;

				//> Prepare collage for positive learning
				int collagePicToShow = 20;
				collage = Mat::zeros(firstFrame.size().width/collagePicToShow,firstFrame.size().width,CV_8UC3);
				collageIndex = 0;
				collageToAddSize = collage.size().height;
				collageRows = collage.size().width / collageToAddSize;
			}


			// Learn offline if present
			if (frameCount==1) {
				
				// Positive
				for(auto& positiveFile : offlinePositiveSamples) {
					Mat train;
					loadImage(positiveFile,train);

					if (newScaleFactor>1) {
						resize(train, train, Size(), newScaleFactor, newScaleFactor, CV_INTER_LINEAR);
					}


					if (train.rows<minObjectSide || train.cols<minObjectSide)
						Log::fatal("Training image ["+positiveFile+"] is too small (minObjectSide: "+toString(minObjectSide)+")");
			
					if (!this->addTrainingSample(train))
						Log::fatal("Training image ["+positiveFile+"] could not be trained. Maybe no keypoints were detected");

					this->configPruningStartFrom++;
				}

				// Negative
				for(auto& negativeFile : offlineNegativeSamples) {
					
					Mat train;
					loadImage(negativeFile,train);

					if (newScaleFactor>1) {
						resize(train, train, Size(), newScaleFactor, newScaleFactor, CV_INTER_LINEAR);
					}


					if (train.rows<minObjectSide || train.cols<minObjectSide)
						Log::fatal("Training image ["+negativeFile+"] is too small (minObjectSide: "+toString(minObjectSide)+")");
			
					if (!this->addNegativeROI(train))
						Log::fatal("Training image ["+negativeFile+"] could not be trained. Maybe no keypoints were detected");

					this->configPruningStartFrom++;
				}

			}


		}

		//> ============ END INITIALIZATION METHODS 

		//> Add a training image
		bool addTrainingSample(const Mat& sample, bool isNegative = false) {

			if (sample.cols < minObjectSide || sample.rows < minObjectSide) {
				Log::warning("[addTrainingSample] Width or Height are less than [minObjectSide:"+toString(minObjectSide)+"], ignoring");
				return false;
			}

			bool isPositive = !isNegative;	//> Simple variable

		
			Log(2)<<"\nLearning "+toString(trains.size()+1)+"th sample " + (isNegative ? "(negative)" : "(positive)");

			vector<vector<KeyPoint>> allKP;
			vector<Mat>	allDesc;


			//> Extract keypoints for each engine registered
			for(unsigned int j=0;j<engine.size();j++) {

			
				vector<KeyPoint> objectKPCurrent;
				Mat objectDescCurrent;

				this->feature(engine[j],sample,objectKPCurrent,objectDescCurrent);

				auto objectKPCount = objectKPCurrent.size();

				if (objectKPCurrent.empty()) {
					//log("No keypoint error for training image. Descriptors size ["+toString(objectDescCurrent.size())+"] image size ["+toString(sample.size())+"]");

					//showImage(sample,"Bad Training");
					//waitKey(0);
					return false;
				}

			
				//> We should erase all keypoints found on negative samples that:
				//> may refer to positive samples 
				//> are already present in the database
				//> Currently this check is disabled. We just learn all negative keypoints without further filtering
				int positiveKeypointDiscarded = 0;
				/*if (isNegative && false) {
					float minDist = engine[j].minAllowedDistance;
					vector<KeyPoint> filteredKP;
					Mat filteredDesc;
					vector<vector<DMatch>> matches;
					engine[j].matcher->knnMatch(objectDescCurrent,matches,2);
					for(auto i=0;i<matches.size();i++) {
						if (matches[i].size()>1) {

							//cout << "Distance allowed: "<<engine[j].minAllowedDistance<<": "<<matches[i][0].distance<<endl;

							//> Very close to a keypoint already present: don't insert it again
							if (matches[i][0].distance > minDist) {
						
								auto qidx = matches[i][0].queryIdx;

								//> Good keypoints, readd
								filteredKP.push_back(objectKPCurrent[qidx]);
								filteredDesc.push_back(objectDescCurrent.row(qidx));

								continue;
							} 
						}
						positiveKeypointDiscarded++;
					} //> loop over matches


					//> Swap
					objectKPCurrent.swap(filteredKP);
					objectDescCurrent = filteredDesc;

					log("Discarded " + toString(positiveKeypointDiscarded) + " keypoints already present from negative (distance less than "+toString(minDist)+")");

				}	//> isNegative*/
			

				//> We should erase all keypoints:	(i) Found on positive samples that may refer to negative keypoints
				//>									(ii) Duplicated of already saved positive keypoints
				int negativeKeypointDiscarded = 0;
				if (isPositive && featureDiscardNegativeWhileLearning) {
				
					vector<KeyPoint> filteredKP;
					Mat filteredDesc;
					vector<DMatch> matches;
					engine[j].matcher->match(objectDescCurrent,matches);

					for(auto i=0;i<matches.size();i++) {

						
						auto trainIndex = matches[i].imgIdx;
						//cout << "Distance: "<<matches[i][0].distance<<(trains[ trainIndex ].isNegative ? " (on negative)":"")<<endl;
						//> If the match was on a negative sample discard it
						//> Or we already have a similar keypoint
						if ( trains[ trainIndex ].isNegative || 
							(matches[i].distance < engine[j].minAllowedDistance) ) {

							negativeKeypointDiscarded++;
							continue;

						} 
					

					
						auto qidx = matches[i].queryIdx;

						//> Good keypoints, readd
						filteredKP.push_back(objectKPCurrent[qidx]);
						filteredDesc.push_back(objectDescCurrent.row(qidx));


					} // for

					Log(2)<<"Number of negative keypoints discarded: "+toString(negativeKeypointDiscarded)+" out of "+
						toString(objectKPCount)+" total matches parsed: " + toString(matches.size());

					//> We need the check because the first training images doens't match anything.
					if ( matches.size()>0 && negativeKeypointDiscarded > 0 ) {
				
						//> If we discarded every keypoints 
						if ( negativeKeypointDiscarded >= objectKPCount ) { 	
							objectKPCurrent.clear();
							Mat emptyMat(0,objectDescCurrent.cols,objectDescCurrent.type());
							objectDescCurrent = emptyMat;
						} else { //> If we discarded only some keypoints

						
							filteredDesc.reshape(0,objectDescCurrent.cols); //> (seems not necessary)

							if (drawTrainingImage) {
								//> Only for debugging
								Mat drawingIm;
								sample.copyTo(drawingIm);
								drawKeypoints(drawingIm,objectKPCurrent,drawingIm,Scalar(0,0,0));
								showImage(drawingIm,toString(engine[j].name)+") Learning, all keypoints");
								drawKeypoints(drawingIm,filteredKP,drawingIm,engine[j].color);
								showImage(drawingIm,toString(engine[j].name)+") Learning, filtered");

							}

							//> Costant time swapping
							objectKPCurrent.swap(filteredKP);
							objectDescCurrent = filteredDesc;
						}
					}

				} //> isNegative


				Log(2)<<"Adding " + toString(objectKPCurrent.size()) + " keypoints";

				//> If there aren't any keypoints (we discarded everything) don't train them
				if ( objectKPCurrent.size()<1 || objectKPCurrent.empty() || objectDescCurrent.empty() ) {
					//log("WARNING: Too few keypoints extracted..");
					//waitKey(500);
					//showImage(sample);

					//if (trains.size()==0)
						//waitKey(0);
				
					return false; //> don't add it!
				}

				allKP.push_back(objectKPCurrent);
				allDesc.push_back(objectDescCurrent);
			
			} //> for engines

			trains.push_back(MultiImage(sample,engine.size()));
			trains.back().isNegative = isNegative;
			trains.back().frameAdded = frameWithObject;
		

			for(auto j=0;j<engine.size();j++) {

				trainTotalKeyPoint += allDesc[j].rows;//> Keep track of the number of keypoints
				if (isNegative)
					trainNegativeKeyPoint += allDesc[j].rows;//> Keep track of the number of keypoints

				trains.back().addFeature(j,allKP[j],allDesc[j]);

				//> Draw keypoints
				if (drawKP)
					drawKeypoints(trains.back().image,allKP[j],trains.back().image,engine[j].color);

				engine[j].matcher->add( trains.back().getEngineDescriptors(j) );
			
			
			}


			if (isNegative) {
				trainNegativeSample++;   //> Increment negative counter

				debugShowImage(trains.back().image, "Learning: Negative");
			} else
				debugShowImage(trains.back().image, "Learning: Positive");

			


			return true;

		} //> addTrainingSample

		//> Add a negative sample
		bool addNegativeSample(Rect roi, const Mat& queryImage) {

			roi = rectSum(roi,configFeatureDiscardNegativeWLenlargeROI);
			rectInsideImage(queryImage,roi);
			
			//> Don't extract negative keypoints inside the bounding box (the positive object)
			detectionMask = Mat(queryImage.size(), CV_8U, Scalar::all(255));
			detectionMask(roi).setTo(Scalar::all(0));

			bool isAdded = this->addNegativeROI(queryImage);

			detectionMask = Mat();

			return isAdded;

		}

		//> Add a negative ROI (the detection mask is already set or the entire image is a negative)
		bool addNegativeROI(const Mat& queryImage) {


			//> Add this negative sample
			if (this->addTrainingSample(queryImage,true)) {

				return true;
			}

			return false;

		}

		//> Add a negative region around the target object
		bool addNegativeAdiacent(const Mat& fullQueryImage) {

			Rect adiacentNegativeROI = rectSum(roi,Point(200,150));
			rectInsideImage(fullQueryImage,adiacentNegativeROI);
			Mat adiacentNegativeImage = fullQueryImage(adiacentNegativeROI);

			
			bool isAdded = this->addNegativeSample(
				Rect(roi.x-adiacentNegativeROI.x,roi.y-adiacentNegativeROI.y,roi.width,roi.height),
				adiacentNegativeImage
			);

			/*if (isAdded && drawKP) {
				//> Debug				
				for(auto e=0;e<engine.size();e++) {
					vector<KeyPoint> k = trains.back().keypoint[e];
					shiftKeyPoint(k,adiacentNegativeROI.tl());
					drawKeypoints(frame,k,frame,Scalar(30,30,30));
				}
			}*/

			return isAdded;
		}

		//> Detect and compute keypoints with the "engine" passed
		//> The function may: (i) wrap the detector with a GridAdapted layer (based on engine.useGrid)
		//>                   (ii) lower the number of keypoints detect based on the image size		
		void feature(const KeyPointEngine& engine,const Mat& object, vector<KeyPoint>& kp, Mat& kpDesc) {

			Log(2)<<"Using: " + engine.name + " on size: " + toString(object.size())<<endl;

			clock_t start = profileStart(1);

			Ptr<FeatureDetector> detector = engine.detector;

			//> If image area is big enough and the engine.useGrid is enabled
			int area = object.size().area();
			if (area > (360*360) && (engine.useGrid)) {


				//> Calc a grid of about 250x250, with 400 keypoint inside
				auto keyPointForCell = configDetectorKeyPointForCell;
				auto gridRows = object.size().height / configDetectorCellSide;
				auto gridCols = object.size().width / configDetectorCellSide;

				if (gridRows<1) gridRows = 1;
				if (gridCols<1) gridCols = 1;

				int totalKeyPointLimit = (gridCols*gridRows) * keyPointForCell;
				Log(2)<<"Uing a GRID. Limit KeyPoint: "+toString(totalKeyPointLimit)<<endl;

				detector = new GridAdaptedFeatureDetector(engine.detector,totalKeyPointLimit,gridRows,gridCols);

				/*if (debug) {
					Mat frameGrid;
					object.copyTo(frameGrid);
					drawGrid(frameGrid,gridRows,gridCols);
					debugShowImage(frameGrid,"Grid Adapted for " + engine.name);
				}*/

			}

			//> SIFT doesn't seem to work with MASK (OpenCV 2.4.5)
			if (!detectionMask.empty() && engine.name != "SIFT") {
				Log(2)<<"Using MASK for detection"<<endl;
				//> Find Keypoints
				detector->detect( object, kp, detectionMask);	
			} else 
				detector->detect( object, kp );	

			//> Manual filtering because mask crashes with SIFT in Debug mode
			if (!detectionMask.empty() && engine.name == "SIFT") {
				Log(2)<<"Detected SIFT, manual mask filtering"<<endl;
				KeyPointsFilter::runByPixelsMask(kp,detectionMask);
			}
			
			start = profileDebug(start,"Detected " + toString(kp.size()) + " kps with " + engine.name);

			//> Further filtering for ORB (returns too many keypoints close to each other)
			if (engine.name == "ORB") {
				filterKeyPointsByKernel(kp,2);
				/*
				Mat debugFilter;
				object.copyTo(debugFilter);
				drawGrid(debugFilter,5);
				drawKeypoints(debugFilter,kp,debugFilter,engine.color);
				showImage(debugFilter,"Filter");*/
			}

			if (!kp.empty()) {
				//Calculate descriptors (feature vectors)
				engine.extractor->compute( object, kp, kpDesc );
				profileDebug(start,"Computed");
			} else {
				Log::warning("No keypoints extracted with " + engine.name);
			}

		} //> ========= END feature()

		//> Matches the descriptors "queryDescCurrent" with the training pool
		void match(
			KeyPointEngine& engine, 
			Mat& queryDescCurrent, 
			vector<DMatch>& goodMatches, 
			vector<KeyPoint>& queryInlierCandidates,
			vector<KeyPoint>& queryKPCurrent
		){
		
			clock_t start = profileStart();

			//> Call the real implementation of the matching 
			this->matchSimpleImplementation(
				engine,queryDescCurrent,goodMatches,queryInlierCandidates,queryKPCurrent
			);
			
			profileDebug(start,"Matching (good match found: "+toString(goodMatches.size())+")");

		}

		//> Default implementaion of the "match" method. Uses NNDR and discard negative keypoints
		//> Note that you can disable the NNDR filter by setting the ratioThreshold of the KeyPointEngine >= 1.0
		void matchDefaultImplementation(	
			KeyPointEngine& engine, 
			Mat& queryDescCurrent, 
			vector<DMatch>& goodMatches, 
			vector<KeyPoint>& queryInlierCandidates,
			vector<KeyPoint>& queryKPCurrent
		){

			vector<vector<DMatch>> matches;
			engine.matcher->knnMatch(queryDescCurrent,matches,2);

			float currentRatio = engine.ratioThreshold;
			int goodMatchCurrentCount = 0;
			int goodNegativeCount = 0;


			Log(2)<<"Parsing " + toString(matches.size()) + " matches"<<endl;
			//> Ratio Test
			for(auto i = 0; i <matches.size(); i++) {


				if (matches[i].size()==0) {
					cout << "WARNING, no matched returned! matches size: "<<matches[i].size()<<endl;
					vector<Mat> td = engine.matcher->getTrainDescriptors();
					for(auto tdi = 0;tdi<td.size();tdi++)
						cout << "Image["<<tdi<<"]: desc size:"<<td[tdi].size()<<endl;

					waitKey(0);
				}
				if (matches[i].size()>1) {
					//cout << "Matches i: "<<i<<" ratio: "<<matches[i][0].distance / matches[i][1].distance<<endl;
					if( (matches[i][0].distance / matches[i][1].distance) <= currentRatio ){

						int idx1 = matches[i][0].imgIdx;
						int idx2 = matches[i][1].imgIdx;
						int goodIdx = 0;

						//> Discard keypoint if they are matched on a negative sample
						if ( trains[idx1].isNegative ) {

							trains[idx1].incrementUsage(); //> Negative usage!

							if (drawKPNegativeDiscarded)
								drawMark(frame,queryKPCurrent[ matches[i][0].queryIdx ]);

							goodNegativeCount++;
							continue;
						}

			
						goodMatchCurrentCount++;
						goodMatches.push_back(matches[i][goodIdx]);

						//> Adding current match to list
						//trains[idx].addGoodMatch(engine.idx,matches[i][0]);

						//> Add the query keypoint to the list, used for drawing
						queryInlierCandidates.push_back( queryKPCurrent[ matches[i][goodIdx].queryIdx ]);
					}
				}
			}

			Log(2)<<"Good match discarded (on negative samples): " + toString(goodNegativeCount)<<endl;
		}

		//> Due to a bug we can't use knnMatch when there is an image with only one descriptor.
		//> This implementation jsut discards negative keypoints
		void matchSimpleImplementation(	
			KeyPointEngine& engine, 
			Mat& queryDescCurrent, 
			vector<DMatch>& goodMatches, 
			vector<KeyPoint>& queryInlierCandidates,
			vector<KeyPoint>& queryKPCurrent
		){

			vector<DMatch> matches;
			engine.matcher->match(queryDescCurrent,matches);

			int goodMatchCurrentCount = 0;
			int goodNegativeCount = 0;


			Log(2)<<"Parsing " + toString(matches.size()) + " matches"<<endl;

			for(auto i = 0; i <matches.size(); i++) {

				int idx1 = matches[i].imgIdx;

				//> Discard keypoint if they are matched on a negative sample
				if ( trains[idx1].isNegative ) {

					trains[idx1].incrementUsage(); //> Negative usage!

					if (drawKPNegativeDiscarded)
						drawMark(frame,queryKPCurrent[ matches[i].queryIdx ]);

					goodNegativeCount++;
					continue;
				}

			
				goodMatchCurrentCount++;
				goodMatches.push_back(matches[i]);

				//> Adding current match to list
				//trains[idx].addGoodMatch(engine.idx,matches[i][0]);

				//> Add the QUERY keypoint to the list, used for drawing
				queryInlierCandidates.push_back( queryKPCurrent[ matches[i].queryIdx ]);
			
			}

			Log(2)<<"Good match discarded (on negative samples): " + toString(goodNegativeCount)<<endl;
		}


		//> Learn new negative/positive samples
		void onlineLearning(const Mat& fullQueryImage,const Mat& currentObjectROI,int fullROIArea) {

			//> ============ Online Positive Learning ==============
			//> ----------------------------------------------------

				bool currentFeatureDiscardNegative = featureDiscardNegativeWhileLearning;

				int step = configLearningPositiveStep;						//> Assign default step
				if (!isObjectFound ||							
					objectHoughBucket.votes < 2*configHoughMinVotes ||		//> Adaptive learning: learn only when necessary
					frameCount<25) { 

					step = 1;	//> If one condition above is verified learn a new positive now
				}


				if (featureOnlinePositiveLearning && 
					fullROIArea == currentObjectROI.size().area() &&				//> Avoid to learn a new positive sample if part of the object is out of scene
					frameCount%step==0)							//> Right frame number
				{	
							
					//> If these are first frames never discard keypoints
					//if (frameCount<15)
						//featureDiscardNegativeWhileLearning = false;

					if (this->addTrainingSample(currentObjectROI)) {

						//imwrite(outputPath+toString(trains.size())+".png",currentObjectROI);

						debugShowImage(trains.back().image,"Learning: Positive");
						addPicToCollage(currentObjectROI);

					} else {
						Log(2) << "Could not learn this positive sample"<<endl;
					}


					featureDiscardNegativeWhileLearning = currentFeatureDiscardNegative;

				} 
			//> ========= End online positive learning



			//>	========== Online negative learning ================
			//> ----------------------------------------------------
				if (featureOnlineNegativeLearning) {

					if ( consecutiveGoodFrames%configLearningNegativeAdiacentStep==0 || frameCount == 1)  {

						addNegativeAdiacent(fullQueryImage);
						Log(2) << "Adiacent Negative learning..."<<endl;
					} 	

					if (consecutiveGoodFrames%configLearningNegativeStep==0 /* && areNegativesWithinLimit() */) {

						this->addNegativeSample(roi, fullQueryImage);
						//waitKey(0);
					}

				}
			//> ============================ End online negative learning

		} //> ====== End online learning

		//> Check if the queryCenter is too far from the previous (with maxObjectShift)
		bool isObjectCloseEnough(const LHTBucket& best, Mat& currentObjectROI) {

			if (frameWithObject > 0 && featureDiscardROITooFar) { //> We found at least once the object
						
				float currentMaxAllowedShift = (frameCount - previousGoodFrame) * maxObjectShift; //> The max allowed shift
				float currentShift = euclideanDist(best.meanQueryCenter,objectHoughBucket.meanQueryCenter);

				Log(2) << "Current object center shift: " + toString(currentShift) + " allowed: " + toString(currentMaxAllowedShift)<<endl;
				if ( currentShift > currentMaxAllowedShift ) {

					//> This ROI can't be the good object!
					//> We should learn this ROI as a negative sample!

					Log(2) << "The current target object is too far from the previous on. Distance: " + toString(currentShift) + " (allowed:" + toString(currentMaxAllowedShift) + ")"<<endl;

					addNegativeROI( currentObjectROI );			
					debugShowImage(currentObjectROI,"Object Too Far");

					return false;
				}
			}

			return true;
		}

		//> When pruning we need to re-build the index of the training pool and 
		//> sync them with the descriptors in the Matcher module
		void reTrain() {
			trainNegativeKeyPoint = 0;
			trainTotalKeyPoint = 0;
			trainNegativeSample = 0;
			trainPositiveKeySample = 0;

			vector<vector<Mat>> readding(engine.size());
		
			for(auto i=0;i<trains.size();i++) {

				//> STAT
					if (trains[i].isNegative)								//> Increment negative counter
						trainNegativeSample++;
					if (!trains[i].isNegative && trains[i].keyTraining)		//> Increment positive key counter
						trainPositiveKeySample++;
				//> =====

				for(auto e=0;e<engine.size();e++) {
					readding[e].push_back( trains[i].desc[e] );

					//> STAT
					trainTotalKeyPoint += trains[i].desc[e].rows;					//> keep track of keypoints
																					//> IT could be used keypoint[e].size(); too
					if (trains[i].isNegative)
						trainNegativeKeyPoint += trains[i].desc[e].rows;		//> keep track of keypoints
					//> ========


				}					
			}

			for(auto e=0;e<engine.size();e++) {
				engine[e].matcher->clear();
				engine[e].matcher->add( readding[e] );

				engine[e].matcher->train();	//> Empty for brute force
			}
		}

		//> Try to balance the ratio of positive/negative keypoints (based on configLearningRatio)
		bool pruneNegative() {

			if (trainNegativeSample<2)
				return false;

			//> Avoid pruning too often
			if (frameCount-lastNegativePruning<25)
				return false;

			lastNegativePruning = frameCount;

			vector<MultiImage> keepSample;
			auto factor = 0.3;	//> Delete first 30%
			auto negativeToKill = ceil(trainNegativeSample * factor);
			auto prunedNegative = 0;

			for(size_t i = 0; i < trains.size(); ++i) {
				if ( !trains[i].isNegative || prunedNegative == negativeToKill || i==1 ) {
					keepSample.push_back(trains[i]);
					continue;
				} else {
					prunedNegative++;
				}
			}

			Log(2) << "Pruned " + toString(prunedNegative)<<endl;

			if (prunedNegative>0) {
				keepSample.swap(trains);
				this->reTrain();
				return true;
			} else 
				return false;

		} //> pruneNegative

		//> Delete oldest positive and negative samples
		void prune() {

		
			auto prunedPositive = 0;	//> Samples
			auto prunedNegative = 0;	//> Samples

			auto prunedPositiveKP = 0;	//> KeyPoints
			auto prunedNegativeKP = 0;	//> KeyPoints

			vector<MultiImage> keepSample;

			auto factor = 0.3;	//> Delete first 30%

			auto trainPositiveKeyPoint = trainTotalKeyPoint - trainNegativeKeyPoint;


			auto positiveKPToPrune = roundToNearest( trainPositiveKeyPoint * factor );
			auto negativeKPToPrune = roundToNearest( trainNegativeKeyPoint * factor );


			for(size_t i = 0; i < trains.size(); ++i) {

				//> Always save first samples
				if (i<configPruningStartFrom) {
					keepSample.push_back(trains[i]);
					continue;
				}

				//> Negative sample
				if ( trains[i].isNegative ) { 

					if ( prunedNegativeKP >= negativeKPToPrune ) //> Save this sample if we already deleted the right amount of keypoints
						keepSample.push_back(trains[i]);
					else {
						prunedNegative++;

						for(auto e=0;e<engine.size();++e)
							prunedNegativeKP += trains[i].desc[e].rows;

					}
						
				} 
				//> Positive sample
				else {
					
					if ( prunedPositiveKP >= positiveKPToPrune ) //> Save this sample if we already deleted the right amount of keypoints
						keepSample.push_back(trains[i]);
					else {
						prunedPositive++;

						for(auto e=0;e<engine.size();++e)
							prunedPositiveKP += (int)trains[i].desc[e].rows;
					}

				}

			}	//> loop over trains

			keepSample.swap(trains);
			this->reTrain();

		}
		
		//> Return true if the % of negative samples is less than configFeatureLearningRatio
		bool areNegativesWithinLimit() {
			return ((float)trainNegativeKeyPoint / trainTotalKeyPoint) < configFeatureLearningRatio;
		}

		//> Calc the target object size with ratio of pairs of keypoints. (See paper for further details)
		//> Big scale changes will be discarded (because with online learning we should never find big scale changes)
		Size2f calcScale(LHTBucket& bucket) {

			vector<Size2f> objectSize;

			for( auto it = bucket.imgMatches.begin(); it != bucket.imgMatches.end(); ++it ) {
				auto matches = bucket.imgMatches[it->first];

				if (debugScale)
					cout << endl <<"Parsing image: "<<it->first<<", size: "<<trains[it->first].image.size()<<", with: "<<matches[0].size()<<" matches";

				//> If there are at least 2 Matches
				if ( matches[0].size() > 1) {
			
					float meanScale = 0;
					int counter = 0;

					for(auto i=0;i<matches[0].size()-1;i++) {

						auto trainDist = euclideanDist( 
							matches[0][i],
							matches[0][i+1]
						);
			
						auto queryDist = euclideanDist( 
							matches[1][i],
							matches[1][i+1]
						);

						//> Discard unlikely points (too close from each other)
						if (trainDist<4 || queryDist<4)
							continue;

			
						float scale = queryDist/trainDist;

						//> Discard huge scale variation
						if (scale <0.4 || scale>1.4 ) //> False positive match ?
							continue;

						meanScale += scale;
						counter++;	

						if (debugScale) {
	
							Mat debugTrain,debugFrame;
							trains[it->first].image.copyTo(debugTrain);
							frame.copyTo(debugFrame);

							drawCircle(debugTrain,matches[0][i]);
							drawCircle(debugTrain,matches[0][i+1]);

							drawCircle(debugFrame,matches[1][i]);
							drawCircle(debugFrame,matches[1][i+1]);

							showImage(debugTrain,"train");
							showImage(debugFrame,"query");

							waitKey(10);
							cout << endl << "Image["<<it->first<<"] Scale factor ("<<scale<<") ";
					
							keyboardCallback(0);
						}
					}


					//> If we used pair of correspondences
					if (counter) {
						meanScale /= counter;

						objectSize.push_back(Size2f(
							trains[it->first].image.size().width*meanScale,
							trains[it->first].image.size().height*meanScale
						));
					}

					if (debugScale) {
						cout <<endl<< " Mean Scale: "<<meanScale<<endl;
						if (meanScale)
							cout <<" Scaled size: "<<objectSize.back()<<endl;
					}

				}
		
			} //> End for (loop over pairs of keypoints)

			Size2f oldObjectSize = bucket.meanTrainSize;
			if (!objectSize.empty()) {	//> Change bucket.mainTrainSize only if we added something
				Size2f meanSize;
				for(auto i=0;i<objectSize.size();++i)
					meanSize += objectSize[i];
	
				meanSize.width /= objectSize.size();
				meanSize.height /= objectSize.size();

				bucket.meanTrainSize = meanSize;

			}

			float meanScale = bucket.meanTrainSize.width / oldObjectSize.width;

			Log(2) << "Final Size: " + toString(bucket.meanTrainSize) + " with count: " + toString(objectSize.size()) + "\n"<<endl;

			return Size2f(meanScale,meanScale);
		}

		//> Add to the technique pool a new keypoint-based method
		void addEngine(
			   Ptr<FeatureDetector> d,
			   Ptr<DescriptorExtractor> e,
			   Ptr<DescriptorMatcher> m,
			   string name,
			   Scalar c,
			   float r = 0.95f,		//> Threshold used for NNDR (we suggest a pretty high value, as outlier will be discarded by Hough transform)
									//> You can let our GHT to take care of every keypoints by disabling NNDR with a ratio r = 1.0 
			   bool g=false			//> Allows the use of GridAdapted layer to detect keypoints
		) {

			engine.push_back(KeyPointEngine(
				d,
				e,
				m,
				name,
				c,
				r,
				g
			));

			auto idx = engine.size()-1;
			engine[ idx ].idx = idx;
		}


		//> =========================================================================
		//>                       Generic helper/debug methods 
		//> =========================================================================

		//> Adds image "toAdd" to the current collage
		void addPicToCollage(const Mat& toAdd) {


			if (collageIndex==collageRows) {
				collageIndex=0;
				collage=Scalar(0);	//> Darken the entire collage section, start over
			}

			Mat toAddCopy;
			toAdd.copyTo(toAddCopy);

			resize(toAdd,toAddCopy,Size(collageToAddSize,collageToAddSize));
			cvtColor(toAddCopy,toAddCopy,CV_GRAY2RGB);

			toAddCopy.copyTo( collage( Rect( collageIndex*collageToAddSize,0,collageToAddSize,collageToAddSize ) ));

			++collageIndex;
		}


		
		//> =============== Debug methods ===================

		
		//> Returns the processor time if the passed level is higher than debugLevel
		clock_t profileStart(int level = 1) {
			if (level > debugLevel)
				return 0;

			return clock();
		}
		//> Returns the current clock_t processor time
		clock_t profileDebug(clock_t start, String label, int level = 1) {
			if (level > debugLevel)
				return 0;

			clock_t current = clock();
			cout << label << ") Time elapsed: " << ((double)current - start) / CLOCKS_PER_SEC << endl;
			return current;
		}
		//> Returns delta in seconds
		double profileDebug(clock_t start) {	
			clock_t current = clock();
			return ((double)current - start) / CLOCKS_PER_SEC;
		}

		//> Prints useful information about the training pool 
		//> When writeTo is different from "" each image in the training pool will be written in the writeTo path
		void debugTrains(bool all = true, string writeTo = "") {

			cout << endl << endl << "= Training image usage: [imageIdx]=>[count]"<<endl;

			auto totalUsage = 0;
			auto keyCount = 0;
	
			auto negCount = 0;


			size_t currentTotalKP= 0;
			size_t currentNegKP= 0;

			for(auto i=0;i<trains.size();i++) {

				auto& t = trains[i];


				for(auto e=0;e<engine.size();e++) {

					currentTotalKP+=t.keypoint[e].size();
					if (t.isNegative)
						currentNegKP+=t.keypoint[e].size();
				}

				if (! (trains[i].isNegative) ) 
					totalUsage+=trains[i].usage;
				else
					negCount++;

				//> If writeTo is a folder write this image inside it
				//if (writeTo.length())
					//imwrite(outputPath + writeTo + "/" + (t.isNegative ? "negative":"positive") +(t.keyTraining ? "_key_":"")+ toString(i) +".png",t.image);
			
				if (i>100 && !all) {
					if (i==101)
						cout << "..."<<endl;
					continue;			
				}


				cout <<"ImageIdx: "<<i<<" used "<<trains[i].usage<<" times "<<(trains[i].isNegative ? "(negative) ":"");

				if (trains[i].keyTraining) {
					cout << "(KEY) ";
					keyCount++;
				}

				if (trains[i].name.size())
					cout << "("<<trains[i].name<<") ";



				cout << "added at: "<<trains[i].frameAdded;


				cout << endl;
			}

			cout << "Total training images: "<<trains.size()<<" (positive key: "<<trainPositiveKeySample<<") with negative images: "<<negCount<<endl;
			cout << "Total keypoint found: "<<currentTotalKP<<", negatives: "<<currentNegKP<<endl;

			cout <<endl<< "Total frame covered: "<<totalUsage<<endl;
			cout << "Total key images " <<keyCount<<endl<<endl;


			waitKey(0);
		}

		//> Print stats about training pool
		void debugTrainingPool() {

			//> Debug information about training pool
			auto totalPositiveKeypoint = trainTotalKeyPoint - trainNegativeKeyPoint;
			Log(1)<<"\nTraining images: " + toString(trains.size()) + " with " + toString(trainNegativeSample) + " negative. Positive Key: "+toString(trainPositiveKeySample)<<endl;
			Log(1) << "Total keypoints learned (limit: " + toString(configFeaturePruningWith) + "): " + toString(trainTotalKeyPoint) + " with " + toString(totalPositiveKeypoint) + " positive and " + toString(trainNegativeKeyPoint) + " negatives"<<endl;
			Log(1) << "Ratio Negative/Total: " + toString((float)trainNegativeKeyPoint / trainTotalKeyPoint) + ", configRatio: " + toString(configFeatureLearningRatio) + "\n"<<endl;

		}

		//> if .debug == true shows im
		void debugShowImage(const Mat& im, const string& name) {
			if (debug) {
				showImage(im,name);
			}
		}


		//> Notify a setting change 
		void cset(string s){
			Log(1) << "Setting [" + s + "]"<<endl;
			waitKey(500);
		}

		//> Manages keyboards input after each frame
		bool keyboardCallback(int delay = 1) {
			char key = waitKey(delay);
			//> Flow
				if (key=='p') { 
					cout <<endl<<"======"<<endl<< "Paused" << endl;
					waitKey();
				} else if (key == 27) {
					return false;
				} 
		
			//> Fallback
				else if (key == 'f') {
					featureFallback = !featureFallback;
					cset("Intelligent Fallback:" + toString(featureFallback));
				
				}
		
			//> Debug
				else if (key == '\'') {
					debugLevel -= 1;
					cset("Lowering Debug level:"+toString(debugLevel));
					if (debugLevel<-1)
						debugLevel = -1;
				} else if (key == '') {
					debugLevel += 1;
					cset("Increasing Debug level:"+toString(debugLevel));
					if (debugLevel>3)
						debugLevel = 3;

				} 
			//> Debug Toggle
				else if (key=='d') {

					setDebug(!debug);
					cset("Toggle debug, now: " +toString(debug));

				}

			//> Debug scale
				else if (key=='s') {
					debugScale = !debugScale;
					cset("Debug Scale:"+toString(debugScale));
				}
			//> Debug trains
				else if (key== 't' ) {
					debugTrains(true,"now");
					waitKey(0);
				}

			//> General
				else if (key == 'k') {
					drawKP = !drawKP;

					cset("Toggle DrawKeyPoint");
				}
				else if (key == 'j') {
					drawKPNegativeDiscarded = !drawKPNegativeDiscarded;

					cset("Toggle drawKPNegativeDiscarded");
				}
				else if (key == 'h') {
					drawHoughVotes = !drawHoughVotes;
					cset("Toggle Hough Debug");
				}

			return true;
		}


};
